from numpy import float32, expand_dims, array
from skimage.io import imread, imsave
from ants import from_numpy, registration, apply_transforms
from re import sub
from os.path import basename, dirname
from tifffile import TiffFile
from tqdm import tqdm
from os.path import exists
from novas3d.utilities import get_mov_files

def register_paired_images(fix_file, mov_files, out_dir, in_filename_extension='.tif', final_filename_extension='_warped.tif', sigma=2, flip = False):
    """
    Register paired images using ANTs registration.

    Args:
        fix_file (str): Path to the fixed image file.
        mov_files (list): List of paths to the moving image files.
        out_dir (str): Output directory to save the registered images.
        in_filename_extension (str, optional): Initial filename extension of the unregistered images. Defaults to '.tif'.
        final_filename_extension (str, optional): Final filename extension of the registered images. Defaults to '_warped.tif'.
        sigma (float, optional): Sigma value for the registration. Defaults to 2.
        flip (bool, optional): Whether to flip the moving image. Defaults to False.

    Returns:
        None
        
    """
    
    # Read the fixed image
    fix_numpy = imread(fix_file)
    fix = from_numpy(float32(fix_numpy[:,0])) # Convert images to ants 

    if flip == True:
        # Extract metadata from the fixed image
        with TiffFile(fix_file) as tif:
            tif_tags = {}
            for tag in tif.pages[0].tags.values():
                name, value = tag.name, tag.value
                tif_tags[name] = value
            image = tif.pages[0].asarray()
        #print(tif_tags)
        start_str = float([x for x in tif_tags['IJMetadata']['Info'].split('\n') if 'axis startPosition #1' in x][0].split(' ')[-1])
        end_str = float([x for x in tif_tags['IJMetadata']['Info'].split('\n') if 'axis endPosition #1' in x][0].split(' ')[-1])
        direction = end_str - start_str
        if [x for x in tif_tags['IJMetadata']['Info'].split('\n') if 'imagingParam zDriveUnitType' in x][0].split(' ')[-1] == 'Piezo':
            direction = float([x for x in tif_tags['IJMetadata']['Info'].split('\n') if 'acquisitionValue zPosition' in x][0].split(' ')[-1])
            piezo = True
        else:
            piezo = False
        #print(direction)
        
    res2 = []
    for mov_file in mov_files:
        # Read the moving image
        mov_numpy = imread(mov_file)
        
        if flip == True and piezo == False:
            with TiffFile(mov_file) as tif:
                tif_tags = {}
                for tag in tif.pages[0].tags.values():
                    name, value = tag.name, tag.value
                    tif_tags[name] = value
                image = tif.pages[0].asarray()
            start_str = float([x for x in tif_tags['IJMetadata']['Info'].split('\n') if 'axis startPosition #1' in x][0].split(' ')[-1])
            end_str = float([x for x in tif_tags['IJMetadata']['Info'].split('\n') if 'axis endPosition #1' in x][0].split(' ')[-1])
            # Flip the moving image if necessary
            if direction * (end_str - start_str) < 0:
                mov_numpy = mov_numpy[::-1]

        elif flip == True and piezo == True:
            with TiffFile(mov_file) as tif:
                tif_tags = {}
                for tag in tif.pages[0].tags.values():
                    name, value = tag.name, tag.value
                    tif_tags[name] = value
                image = tif.pages[0].asarray()
            piezo_dir = float([x for x in tif_tags['IJMetadata']['Info'].split('\n') if 'acquisitionValue zPosition' in x][0].split(' ')[-1])
            # Flip the moving image if necessary
            if direction != piezo_dir:
                mov_numpy = mov_numpy[::-1]

        mov = from_numpy(float32(mov_numpy[:,0]))

        # Register images and get displacement
        mytx = registration(fixed=fix,
                            moving=mov,
                            type_of_transform='Rigid',
                            total_sigma=sigma,
                            aff_metric='meansquares')

        # Move vascular channel
        warpedraw_1 = apply_transforms(fixed=fix,
                                       moving=from_numpy(float32(mov_numpy[:,0])),
                                       transformlist=mytx['fwdtransforms'],
                                       interpolator='linear').numpy()

        # Move neuron channel
        warpedraw_2 = apply_transforms(fixed=fix,
                                       moving=from_numpy(float32(mov_numpy[:,1])),
                                       transformlist=mytx['fwdtransforms'],
                                       interpolator='linear').numpy()
        
        # Combine moved channels into one image
        mov_numpy[:,0,:,:] = warpedraw_1[:,:,:]
        mov_numpy[:,1,:,:] = warpedraw_2[:,:,:]

        # Save warped followup image and baseline image
        imsave(out_dir + sub(in_filename_extension,final_filename_extension,basename(dirname(mov_file)) + '-' + basename(mov_file)),mov_numpy)

    # Save the fixed image
    imsave(out_dir + sub(in_filename_extension,final_filename_extension,basename(dirname(fix_file)) + '-' + basename(fix_file)),fix_numpy)

class ImageRegistration:

    """
    Register images using ANTs registration. Save the registered images to an output directory.
    
    Args:
        images (list): List of image files to register. Each timepoint should be its own file. Files should be tif stacks. Channel 1 should be the color channel. 
        out_directory (str): Output directory to save the registered images.
        in_filename_extension (str): Initial filename extension of the unregistered images.
        final_filename_extension (str): Final filename extension of the registered images.
        timepoint_suffixes (list): List of timepoint suffixes.
        sigma (float): Sigma value for the registration.
        flip (bool): Whether to flip the moving image based on metadata. (only work on images aquired with olympus systems, set to false if not sure if images were aquired on an olympus microscope)
        dic (dictionary): dictionary of the image files with a reference image as the key and a list of moving images as the value.
        skip_existing (bool): Whether to skip already registered images.
    
    Returns:
        None
    """

    def __init__(self, images, out_directory, in_filename_extension, final_filename_extension, timepoint_suffixes, sigma, flip, dic, skip_existing=False):
        self.images = images
        self.out_directory = out_directory
        self.in_filename_extension = in_filename_extension
        self.final_filename_extension = final_filename_extension
        self.timepoint_suffixes = timepoint_suffixes
        self.sigma = sigma
        self.flip = flip
        self.dic = dic
        self.skip_existing = skip_existing

    def register_images(self):
        for image in tqdm(self.images):
            fix_file = sub('_\d{4}', '', image)
            warped_file = basename(dirname(fix_file)) + '-' + basename(fix_file)
            warped_file_path = self.out_directory + sub(self.in_filename_extension, self.final_filename_extension, warped_file)
                
            if self.skip_existing==False or not exists(warped_file_path):
                mov_files = get_mov_files(image, self.dic, self.in_filename_extension, self.timepoint_suffixes)

                register_paired_images(fix_file,
                                       mov_files,
                                       self.out_directory,
                                       self.in_filename_extension,
                                       self.final_filename_extension,
                                       self.sigma,
                                       self.flip)
    


from cc3d import connected_components
from numpy import unique, delete, zeros, isin, hstack, asarray, argmin, array, where
from numpy import min as npmin
from numpy import sum as npsum
from scipy.spatial.distance import cdist
from vg import normalize, perpendicular, angle, basis
from scipy.spatial.transform import Rotation
from pytransform3d.rotations import matrix_from_axis_angle
from re import sub
from numpy import unique, copy
from os.path import exists

def remove_small_comps_3d(image, thresh=500):
    """
    Remove small connected components from a 3D image.

    Parameters:
        image (ndarray): The input 3D image.
        thresh (int): The threshold value to determine the minimum size of connected components to keep.

    Returns:
        ndarray: The filtered image with small connected components removed.
    """
    # Convert the input image to a labeled image where each connected component is assigned a unique label
    img_lab, N = connected_components(image, return_N=True)

    # Count the number of pixels for each unique label
    _unique, counts = unique(img_lab, return_counts=True)

    # Keep only the labels that have pixel count greater than the threshold
    unique_keep = _unique[counts > thresh]

    # Remove the background label (label 0)
    unique_keep = delete(unique_keep, [0])

    # Create a filtered image where only the connected components with labels in unique_keep are kept
    img_filt = zeros(img_lab.shape).astype('int8')
    img_filt[isin(img_lab, unique_keep)] = 1

    return img_filt.astype('uint8')

def fill_holes(img, thresh=1000):
    """
    Fills holes in a binary image using connected component analysis.

    Args:
        img (ndarray): The input binary image.
        thresh (int): The threshold value for removing small components. Default is 1000.

    Returns:
        ndarray: The binary image with filled holes.

    """

    assert len(unique(img)==2), 'Input image must be binary'

    # Iterate over unique values in the image in reverse order
    for i in unique(img)[:]:
    #for i in unique(img)[:]:
        # Create a temporary binary image with only the current value
        _tmp = (img == i) * 1.0
        _tmp = _tmp.astype('int8')
        #print(_tmp)
        
        # Remove small components from the temporary image
        _tmp = remove_small_comps_3d(_tmp, thresh=thresh)
        
        # Update the original image with the filled holes
        img = (_tmp==i)*1 
    
    # Convert the image to int8 type and return
    res = img.astype('int8')
    return res

def _rotmat(vector, points):
    """
    Rotate the given points around the given vector.

    Args:
        vector (array-like): The vector around which the points should be rotated.
        points (array-like): The points to be rotated.

    Returns:
        array-like: The rotated points.
    """
    # Normalize the vector
    vector = normalize(vector)

    # Calculate the axis of rotation
    axis = perpendicular(basis.z, vector)

    # Calculate the angle of rotation in radians
    _angle = angle(basis.z, vector, units='rad')

    # Create the rotation matrix
    a = hstack((axis, (_angle,)))
    R = matrix_from_axis_angle(a)

    # Apply the rotation matrix to the points
    r = Rotation.from_matrix(R)
    rotmat = r.apply(points)

    return rotmat

def closest_node(node, nodes, maximal_distance=10):
    """
    Find the closest node to a given node from a list of nodes.

    Parameters:
        node (array-like): The node for which the closest node needs to be found.
        nodes (array-like): The list of nodes to search from.
        maximal_distance (float): The maximum distance allowed for a node to be considered as the closest. Default is 10.

    Returns:
        array-like: The closest node to the given node from the list of nodes.
    """
    # Convert nodes to a numpy array for efficient computation
    nodes = asarray(nodes)

    # Calculate the squared Euclidean distance between each node and the given node
    dist_2 = npsum((nodes - node)**2, axis=1)

    # Check if the minimum distance is greater than 10
    if len(dist_2) == 0:
        return node
    elif npmin(dist_2) > maximal_distance:
        # If the minimum distance is greater than 10, return the given node itself
        return node
    else:
        # If the minimum distance is less than or equal to 10, return the node with the minimum distance
        return nodes[argmin(dist_2)]

def get_mov_files(image,dic,suffix,timepoint_suffixes = ['_0001']):
    """
    Get a list of files associated with an image.

        Parameters:
            image (str): The image file name.
            dic (dict): A dictionary mapping a key to a list of file names.
            suffix (str): The file name suffix.
            timepoint_suffixes (list): A list of timepoint suffixes. Default is ['_0001'].

        Returns:
            list: A list of files associated with the image.
    """
    # Find the key in the dictionary that matches a substring of the image file name
    key = [x for x in list(dic.keys()) if x in image][0]

    # Generate a list of movie files by replacing the key with each value in the dictionary
    mov_files = [sub(key,x,image) for x in dic[key]]
    mov_files = mov_files + [image]

    # Create a copy of the original list of movie files
    mov_files_init = copy(mov_files)

    # Check if the image file name contains any of the timepoint suffixes
    if any(_suffix in image for _suffix in timepoint_suffixes):
        # If it does, find the specific timepoint suffix
        for x in timepoint_suffixes:
            if x in image:
                _suffix = x

        # Generate a new list of movie files by replacing the timepoint suffix with each value in the timepoint_suffixes list
        for timepoint_suffix in timepoint_suffixes:
            mov_files = sorted(mov_files + [sub(_suffix+suffix,timepoint_suffix + suffix,x) for x in mov_files_init])
            mov_files = sorted(mov_files + [sub(_suffix+suffix,suffix,x) for x in mov_files_init])
        mov_files = mov_files + [sub(_suffix, '',image)]
        #print(mov_files)
    else:
        # If the image file name does not contain any timepoint suffixes, generate a new list of movie files by replacing the suffix with each value in the timepoint_suffixes list
        for timepoint_suffix in timepoint_suffixes:
            mov_files = sorted(mov_files + [sub(suffix,timepoint_suffix + suffix,x) for x in mov_files_init])
            mov_files.append(sub(suffix,timepoint_suffix + suffix,image))



    # Remove the original image file name from the list of movie files
    mov_files = [x for x in mov_files if x != image]

    # Chech to see if the file exists

    mov_files = [x for x in mov_files if exists(x)]

    # Sort the list of movie files
    mov_files = sorted(mov_files)

    # Remove any duplicate movie files
    mov_files = unique(mov_files)

    # Return the final list of movie files
    return mov_files

def remove_terminal_edges(graph, skel, thresh_remove_terminal_segemnts):
    """
    Removes terminal edges from a graph based on a given threshold.

    Parameters:
        graph (networkx.Graph): The input graph.
        skel (numpy.ndarray): The skeleton array.
        thresh_remove_terminal_segemnts (float): The threshold to remove terminal segments.

    Returns:
        numpy.ndarray: The updated skeleton array after removing terminal edges.
    """
    for edge in graph.edges:
        if (graph.degree(edge[0]) == 1 and graph.degree(edge[1]) > 2) or (graph.degree(edge[1]) == 1 and graph.degree(edge[0]) > 2):
            if graph[edge[0]][edge[1]]['weight'] < thresh_remove_terminal_segemnts:
                skel[graph[edge[0]][edge[1]]['pts'][1:-1, 0], graph[edge[0]][edge[1]]['pts'][1:-1, 1], graph[edge[0]][edge[1]]['pts'][1:-1, 2]] = 0
    return skel



from numpy import uint32, int64, save, float16, zeros, load, std, copy, sum, swapaxes, load, mgrid, dstack, array, diag, copy, int16, mean, max, float32, float64, gradient, arange, meshgrid, zeros, vstack, reshape, dot, cos, deg2rad, sin, less, nanargmin, isnan, where, logical_and, sqrt, square,uint8
from numpy.random import shuffle, uniform
from skimage.io import imread, imsave
from pathlib import Path
from tqdm import tqdm
from skimage.morphology import binary_dilation, binary_closing
from scipy.ndimage import distance_transform_edt, zoom, gaussian_filter1d, map_coordinates, gaussian_filter
from scipy.signal import argrelextrema
from networkx import read_gpickle, write_gpickle, selfloop_edges, isolates
from multiprocessing.pool import ThreadPool
from scipy.stats import multivariate_normal
from skimage.restoration import richardson_lucy
from skimage.feature import peak_local_max
from novas3d.utilities import fill_holes, remove_small_comps_3d, get_mov_files, remove_terminal_edges, _rotmat, closest_node
from sknw import build_sknw
from scipy.io import savemat, loadmat
from os.path import exists
from re import sub
from scipy.signal import butter, filtfilt







def save_std_file(file,
                  in_directory='tmp',
                  out_directory='tmp',
                  in_tag='pred',
                  final_tag='std',
                  skip_existing=False):
    """
    Save the standard deviation file.

    Args:
        file (str): The path to the file containing the predicted data.
        in_directory (str): The input directory path.
        out_directory (str): The output directory path.
        in_tag (str): The initial tag for predicted data.
        final_tag (str): The final tag for standard deviation data.

    Returns:
        None
    """

    # Check if the standard deviation file already exists
    if skip_existing==False or not exists(sub(in_directory, out_directory, sub(in_tag, final_tag, file))):
        # Load the predicted data
        pred = load(file)
        # Calculate the standard deviation along the first axis
        _std = std(pred, axis=0)
        # Save the standard deviation file
        save(sub(in_directory, out_directory, sub(in_tag, final_tag, file)), _std)
        


class SaveStdFile:
    """
    Initializes an instance of the SaveStdFile class. Calculates the standard deviation for predicted data. Saves the standard deviation files.

    Args:
        files (list): List of file paths.
        in_directory (str): Input directory path.
        out_directory (str): Output directory path.
        in_tag (str): Initial tag.
        final_tag (str): Final tag.
        skip_existing (bool): Skip files where std already exists.
    
    Methods:
        std(): Save the standard deviation files.  
    """
    def __init__(self, files, in_directory, out_directory, in_tag, final_tag,skip_existing=False):
        self.files = files
        self.in_directory = in_directory
        self.out_directory = out_directory
        self.in_tag = in_tag
        self.final_tag = final_tag
        self.skip_existing = skip_existing

    def std(self):
        """
        Calculate the standard deviation for each file in the files list.

        Returns:
            None
        """
        # Iterate over the files list
        for file in tqdm(self.files):
            # Call the save_std_file function with the file path and other parameters
            save_std_file(file, self.in_directory, self.out_directory, self.in_tag, self.final_tag,self.skip_existing)



class Binarize:
    """
    Initializes the Binarize class. This class is used to binarize the files in the specified directory based on mean and standard deviation thresholds.

    Args:
        directory (str): The directory path.
        min_prob (float): The minimum probability threshold.
        max_var (float): The maximum variance threshold.
        in_directory (str): The input directory path.
        out_directory (str): The output directory path.
        mean_tag (str): The mean tag.
        std_tag (str): The standard deviation tag.
        seg_tag (str): The segmentation tag.
        skip_existing (bool): Skip existing files that have already been binarized.
    
    methods:
        binarize_files(): Binarizes the files in the directory.
        get_files(): Retrieves the list of files in the directory.
    """

    """
    Binarizes the files in the directory.

    This method performs the following steps:
    1. Retrieves the list of files.
    2. Shuffles the list of files.
    3. Iterates over each file.
    4. Checks if the segmented file already exists.
    5. If the segmented file does not exist, checks if the standard deviation file exists.
    6. If the standard deviation file exists, loads the mean and standard deviation files.
    7. Creates an empty segmentation array.
    8. Segments the image based on mean and standard deviation thresholds.
    9. Fills holes in the segmented image.
    10. Removes small connected components from the segmented image.
    11. Saves the segmented image.
    """

    def __init__(self, files, min_prob, max_var, in_directory, out_directory, mean_tag, std_tag, seg_tag, skip_existing=False):
        self.files = files
        self.min_prob = min_prob
        self.max_var = max_var
        self.in_directory = in_directory
        self.out_directory = out_directory
        self.mean_tag = mean_tag
        self.std_tag = std_tag
        self.seg_tag = seg_tag
        self.skip_existing = skip_existing

    def binarize_files(self):
        """
        Binarizes the files in the specified directory based on mean and standard deviation thresholds.

        This method iterates through the files in the directory and performs the following steps for each file:
        1. Checks if the segmented file already exists.
        2. If the segmented file does not exist, checks if the standard deviation file exists.
        3. If the standard deviation file exists, loads the mean and standard deviation arrays from the files.
        4. Creates an empty segmentation array with the same shape as the mean array.
        5. Segments the image based on mean and standard deviation thresholds.
        6. Fills holes in the segmented image.
        7. Removes small connected components from the segmented image.
        8. Saves the segmented image.

        Parameters:
        - None

        Returns:
        - None
        """
        files = self.files
        shuffle(files)

        for file in tqdm(files):
            # Check if the segmented file already exists
            seg_file = sub(self.in_directory, self.out_directory, sub(self.mean_tag, self.seg_tag, file))
            print(seg_file)
            if self.skip_existing==False or not exists(seg_file):
                # Check if the standard deviation file exists
                std_file = sub(self.mean_tag,self.std_tag,file)
                print(std_file)
                if exists(std_file):
                    print(file)
                    mean = load(file)
                    std = load(std_file)
                    seg = zeros(mean.shape[1:]).astype('int8')
                    # Segment the image based on mean and standard deviation thresholds
                    seg[(mean[1,:,:,:] > self.min_prob) * (std[1,:,:,:] < self.max_var)] = 1
                    seg[(mean[2,:,:,:] > self.min_prob) * (std[2,:,:,:] < self.max_var)] = 2
                    seg = seg.astype('int8')
                    seg = (seg==1)*1
                    # Fill holes in the segmented image
                    seg = fill_holes(seg)
                    # Remove small connected components from the segmented image
                    seg = remove_small_comps_3d(seg)
                    print(seg.shape)
                    # Save the segmented image
                    save(seg_file, seg)
    

class NeuronDistanceCalculator:
    """
    Calculates the distance between neurons based on given parameters. Saves the distance transform of the Neuron segmentation array.

    Args:
        files (list): List of file paths.
        min_prob (float): Minimum probability threshold.
        max_var (float): Maximum variance threshold.
        in_directory (str): Input directory path.
        out_directory (str): Output directory path.
        mean_tag (str): Tag for mean values.
        std_tag (str): Tag for standard deviation values.
        neuron_distance_tag (str): Tag for neuron distance values.
        seg_tag (str): Tag for segmentation values.
        seg_nrn_tag (str): Tag for segmented neuron values.
        skip_existing (bool): Whether to skip images that have already been predicted.

    Methods:
        calculate_distance(): Calculates the distance between neurons for each file in the given list.
    """
    def __init__(self, files, min_prob, max_var, in_directory, out_directory, mean_tag, std_tag, neuron_distance_tag, seg_tag, seg_nrn_tag,skip_existing=False):
        self.files = files
        self.min_prob = min_prob
        self.max_var = max_var
        self.in_directory = in_directory
        self.out_directory = out_directory
        self.mean_tag = mean_tag
        self.std_tag = std_tag
        self.neuron_distance_tag = neuron_distance_tag
        self.seg_tag = seg_tag
        self.seg_nrn_tag = seg_nrn_tag
        self.skip_existing = skip_existing

    def calculate_distance(self):
        """
        Calculate the distance between neurons in the given files.

        This method shuffles the list of files and iterates over each file.
        For each file, it checks if the corresponding output file does not exist.
        If it doesn't exist, it loads the mean and standard deviation arrays from the file.
        It then creates a segmentation array based on certain conditions.
        The segmentation array is saved as a binary mask in the output directory.
        Additionally, the distance transform of the segmentation array is calculated and saved as well.

        Args:
            None

        Returns:
            None
        """
        files = self.files
        shuffle(files)

        for file in tqdm(files):
            if not exists(sub(self.in_directory, self.out_directory, sub(self.mean_tag, self.neuron_distance_tag, file))) or self.skip_existing==False:
                if exists(sub(self.mean_tag, self.std_tag, file)):
                    mean = load(file)
                    std = load(sub(self.mean_tag, self.std_tag, file))
                    seg = zeros(mean.shape[1:])
                    seg[(mean[1, :, :, :] > self.min_prob) * (std[1, :, :, :] < self.max_var)] = 1
                    seg[(mean[2, :, :, :] > self.min_prob) * (std[2, :, :, :] < self.max_var)] = 2
                    seg = seg.astype('int8')
                    seg = (seg == 2) * 1
                    save(sub(self.in_directory, self.out_directory, sub(self.mean_tag, self.seg_nrn_tag, file)), seg)
                    save(sub(self.in_directory, self.out_directory, sub(self.mean_tag, self.neuron_distance_tag, file)),
                         distance_transform_edt(1 - seg).astype('float16'))
                    




class SegmentationMaskUnion:
    def __init__(self, files, dic, in_directory, out_directory, in_suffix, final_suffix_mat, final_suffix_npy, timepoint_suffixes, IOU_thresh, thresh_fill_holes, thresh_remove_small_comps, n_interations_binary_closing,skip_existing=False):
        self.files = files
        self.dic = dic
        self.in_directory = in_directory
        self.out_directory = out_directory
        self.in_suffix = in_suffix
        self.final_suffix_mat = final_suffix_mat
        self.final_suffix_npy = final_suffix_npy
        self.timepoint_suffixes = timepoint_suffixes
        self.IOU_thresh = IOU_thresh
        self.thresh_fill_holes = thresh_fill_holes
        self.thresh_remove_small_comps = thresh_remove_small_comps
        self.n_interations_binary_closing = n_interations_binary_closing
        self.skip_existing = skip_existing

    # Define a method called process_images
    def process_images(self):
        """
        Process a list of images by performing binary dilation, calculating intersection over union (IOU),
        performing post-processing, and saving the processed images as .mat and .npy files.

        Args:
            files (list): List of image file paths.
            dic (dict): Dictionary containing additional information.
            in_directory (str): Path to the input directory.
            out_directory (str): Path to the output directory.
            in_suffix (str): Suffix of the initial files.
            final_suffix_mat (str): Suffix of the final .mat files.
            final_suffix_npy (str): Suffix of the final .npy files.
            timepoint_suffixes (list): List of timepoint suffixes. ex: ['_0001', '_0002', '_0003']
            IOU_thresh (float): Intersection over Union (IOU) threshold.
            thresh_fill_holes (int): Threshold for filling holes in the segmented image.
            thresh_remove_small_comps (int): Threshold for removing small connected components in the segmented image.
            n_interations_binary_closing (int): Number of iterations for binary closing.
            skip_existing (bool): Skip existing files that have already been processed.

        Returns:
            None
        """
        # Iterate over each image in the list of files
        for image in tqdm(self.files):
            # Check if the processed image already exists
            if self.skip_existing==False or not exists(sub(self.in_directory, self.out_directory, sub(self.in_suffix, self.final_suffix_mat, sub('_\d{4}', '', image)))):
                # Load the image and perform binary dilation
                img = binary_dilation(load(image))
                _img = copy(img)
                # Get the list of moving files
                mov_files = get_mov_files(image, self.dic, self.in_suffix, self.timepoint_suffixes)
                # Iterate over each moving file
                for i in mov_files:
                    # Load the moving file and perform binary dilation
                    img_0001 = binary_dilation(load(i))
                    # Calculate the overlap and union of the images
                    overlap = (_img != 0) * (img_0001 != 0)  # Logical AND
                    union = (_img != 0) + (img_0001 != 0)  # Logical OR
                    # Calculate the Intersection over Union (IOU)
                    IOU = overlap.sum() / float(union.sum())
                    print(IOU)
                    # Check if the IOU is above the threshold
                    if IOU > self.IOU_thresh:
                        img += img_0001
                # Perform post-processing on the segmented image
                seg = img
                seg[seg != 0] = seg[seg != 0] / seg[seg != 0]
                seg = (seg == 1) * 1
                seg = seg.astype('int8')
                seg = fill_holes(seg, thresh=self.thresh_fill_holes)
                for i in range(self.n_interations_binary_closing):
                    seg = binary_closing(seg)
                seg = remove_small_comps_3d(seg, thresh=self.thresh_remove_small_comps)
                # Save the processed image as a .mat file
                savemat(sub(self.in_directory, self.out_directory ,sub(self.in_suffix, self.final_suffix_mat, sub('_\d{4}', '', image))),
                         {'FinalImage': fill_holes(binary_closing(seg), thresh=self.thresh_fill_holes)})
                # Save the processed image as a .npy file
                save(sub(self.in_directory, self.out_directory, sub(self.in_suffix, self.final_suffix_npy, sub('_\d{4}', '', image))),
                     fill_holes(binary_closing(seg), thresh=self.thresh_fill_holes))

class GraphGenerator:
    """
    Generates graphs based on input files and parameters. Saves the graphs as pickle files. 

    Args:
        files (list): List of input file paths.
        dic (dict): Dictionary containing paried files. Key is the reference file and value is a list of matching files.
        in_directory (str): Path to the input directory.
        out_directory (str): Path to the output directory.
        in_suffix (str): Suffix of the initial files.
        final_suffix_pickle (str): Suffix of the final pickle files.
        final_suffix_mat (str): Suffix of the final mat files.
        final_suffix_tif (str): Suffix of the final tif files.
        ref_timepoint (str): Reference timepoint.
        timepoint_suffixes (list): List of timepoint suffixes.
        IOU_thresh (float): IOU threshold.
        thresh_remove_terminal_segemnts (int): Threshold to remove terminal segments.
        skip_existing (bool): Skip existing files that have already been processed.

    Methods:
        process_files(): Process the input files and generate graphs.

    """

    def __init__(self, files, dic, in_directory, out_directory, in_suffix, final_suffix_pickle, in_suffix_mat, final_suffix_tif, ref_timepoint, timepoint_suffixes, IOU_thresh, thresh_remove_terminal_segemnts, skip_existing=False):
        self.files = files
        self.dic = dic
        self.in_directory = in_directory
        self.out_directory = out_directory
        self.in_suffix = in_suffix
        self.final_suffix_pickle = final_suffix_pickle
        self.in_suffix_mat = in_suffix_mat
        self.final_suffix_tif = final_suffix_tif
        self.ref_timepoint = ref_timepoint
        self.timepoint_suffixes = timepoint_suffixes
        self.IOU_thresh = IOU_thresh
        self.thresh_remove_terminal_segemnts = thresh_remove_terminal_segemnts
        self.skip_existing = skip_existing

    def generate_graphs(self):
        """
        Process the files in the specified directory.

        This method performs the following steps:
        1. Iterate over each file in the directory.
        2. Check if the corresponding output file does not exist.
        3. If the output file does not exist, check if the corresponding input file exists.
        4. If the input file exists, load the skeleton image from the pickle file.
        5. If the skeleton image is not empty, build a graph representation of the skeleton.
        6. Remove self-loop edges from the graph.
        7. Remove terminal edges from the graph.
        8. Save the skeleton image and the graph as pickle files.
        9. For each movement file associated with the current file:
            a. Check if the movement file exists.
            b. If the movement file exists, load the segmented image.
            c. Calculate the intersection over union (IOU) between the segmented images.
            d. If the IOU is above the threshold, save the graph as a pickle file for the movement file.

        Note: This method assumes the existence of certain variables and functions, such as `self.files`, `self.in_directory`, `self.out_directory`, `self.in_suffix`, `self.final_suffix_pickle`, `self.ref_timepoint`, `self.final_suffix_mat`, `self.thresh_remove_terminal_segemnts`, `self.final_suffix_tif`, `self.dic`, `self.timepoint_suffixes`, `self.IOU_thresh`, `loadmat`, `build_sknw`, `selfloop_edges`, `remove_terminal_edges`, `imsave`, `write_gpickle`, `get_mov_files`, `load`, `binary_dilation`, `exists`, `sub`, and `isolates`. Please make sure these variables and functions are defined before calling this method.
        """
        
        for file in tqdm(self.files):
            # Check if the output file already exists
             if self.skip_existing==False or not exists(sub(self.in_directory, self.out_directory, sub(self.in_suffix, self.final_suffix_pickle, file))):
            # Check if the corresponding input file exists
                if exists(sub(self.ref_timepoint + self.in_suffix, self.in_suffix_mat, file)):
                    # Load the skeleton image from the pickle file
                    skel_file = sub(self.ref_timepoint + self.in_suffix, self.in_suffix_mat, file) 
                    skel = loadmat(skel_file)['FilteredImage']
                    skel = skel.astype(int64)
                    # Check if the skeleton image is not empty
                    if sum(skel) != 0:
                        # Build a graph representation of the skeleton
                        graph = build_sknw(skel.astype(uint32), multi=False)
                        print(len(graph.edges))
                        # Remove self-loop edges from the graph
                        while len(list(selfloop_edges(graph))) > 0:
                            if len(list(selfloop_edges(graph))) != 0:
                                for edge in list(selfloop_edges(graph)):
                                    skel[graph[edge[0]][edge[1]]['pts'][1:-1, 0], graph[edge[0]][edge[1]]['pts'][1:-1, 1], graph[edge[0]][edge[1]]['pts'][1:-1, 2]] = 0
                                # Remove terminal edges from the graph
                                skel = remove_terminal_edges(graph, skel, self.thresh_remove_terminal_segemnts)
                                graph = build_sknw(skel.astype(uint32), multi=False)
                        skel = remove_terminal_edges(graph, skel, self.thresh_remove_terminal_segemnts)
                        graph = build_sknw(skel.astype(uint32), multi=False)
                        # Remove isolated nodes from the graph
                        graph.remove_nodes_from(list(isolates(graph)))
                        print(len(graph.edges))
                        # Save the skeleton image as a tif file
                        imsave(sub(self.in_directory, self.out_directory,sub(self.ref_timepoint + self.in_suffix, self.final_suffix_tif, file)), skel)
                        # Save the graph as a pickle file
                        write_gpickle(graph, sub(self.in_directory, self.out_directory, sub(self.in_suffix, self.final_suffix_pickle, file)))
                        # Process movement files associated with the current file
                        print(file)
                        mov_files = get_mov_files(file, self.dic, self.in_suffix, self.timepoint_suffixes)
                        print(mov_files)
                        #seg = load(sub('_\d{4}','',file))
                        seg = load(file)
                        seg = binary_closing(seg)
                        seg = fill_holes(seg)
                        seg = binary_dilation(seg)
                        seg = fill_holes(seg)
                        seg = remove_small_comps_3d(seg)
                        for _file in mov_files:
                            if exists(_file):
                                seg_0001 = load(_file)
                                seg_0001 = binary_closing(seg_0001)
                                seg_0001 = fill_holes(seg_0001)
                                seg_0001 = binary_dilation(seg_0001)
                                seg_0001 = fill_holes(seg_0001)
                                seg_0001 = remove_small_comps_3d(seg_0001)
                                # Calculate the intersection over union (IOU) between the segmented images
                                overlap = (seg != 0) * (seg_0001 != 0)  # Logical AND
                                union = (seg != 0) + (seg_0001 != 0)  # Logical OR
                                IOU = overlap.sum() / float(union.sum())
                                print(IOU)
                                # If the IOU is above the threshold, save the graph as a pickle file for the movement file
                                if IOU > self.IOU_thresh:
                                    write_gpickle(graph, sub(self.in_directory, self.out_directory, sub(self.in_suffix, self.final_suffix_pickle, _file)))
                                else:
                                    print('IOU too low')

class VesselRadiiCalc:

    '''
    Initializes an instance of the VesselRadiiCalc class. Calculates the radii of vessels based on given parameters. Saves the radii in the graph.

    Args:
        files (list): List of file paths.
        in_directory (str): Input directory path.
        img_directory (str): Image directory path.
        mean_directory (str): Mean directory path.
        std_directory (str): Standard deviation directory path.
        out_directory (str): Output directory path.
        pickle_file_suffix (str): Suffix of the pickle files.
        out_pickle_suffix (str): Suffix of the output pickle files.
        img_suffix (str): Suffix of the image files.
        seg_suffix (str): Suffix of the segmented files.
        mean_suffix (str): Suffix of the mean files.
        std_suffix (str): Suffix of the standard deviation files.
        neuron_distance_suffix (str): Suffix of the neuron distance files.
        second_channel (bool): Second channel flag. Denotes the presence of a second channel.
        neuron_channel (bool): Neuron channel flag. Denotes the presence of a neuron channel.
        psf (list): Point spread function.
        spacing (float): Spacing value for pixel size along each axis.
        vessel_segment_limit (int): Vessel segment limit. If larger than the limit, the file is skipped.
        max_pixel_value (int): Maximum pixel value.
        n_iter_deconv (int): Number of iterations for deconvolution.
        grid_size_psf_deconv (int): Grid size for deconvolution. This must be an odd number.
        sampling (float): Sampling in microns to calculate the radii.
        n_cores (int): Number of cores.
        filter_radii (bool): Filter radii with a butterworth filter.
        butter_N (int): Butterworth filter order.
        butter_Wn (float): Butterworth filter frequency.
        butter_btype (str): Butterworth filter type.
        butter_fs (float): Butterworth filter frequency.
        skip_existing (bool): Skip existing files that have already been processed.
    
        return:
            Saves the radii in the graph, along with the standard deviation of the readius estimates along the path.
            If the filter_radii flag is set to True, the radii are filtered with a butterworth filter and saved under the path_weights key in the graph. The unfiltered radii are saved under the path_weights_unfiltered key in the graph.
            If the filter_radii flag is set to False, the radii are saved as is under the path_weights key in the graph.
            In both cases, the mean radii estimate is calculated from unfilterd radii estimates and saved under the radii key in the graph.
            The standard deviation of the radii estimates is calculated and saved under the radii_std key in the graph.
            If neuron_channel is set to True, the mean distance to the closest neuron is calculated and saved under the mean_neuron_distance key in the graph.
            If neuron_channel is set to True, the standard deviation of the distance to the closest neuron is calculated and saved under the neuron_distance_std key in the graph.
            If neuron_channel is set to True, the distance to the closest neuron is saved under the neuron_distance_min key in the graph.

    '''

    def __init__(self, files, in_directory, img_directory, mean_directory, std_directory, out_directory, pickle_file_suffix, out_pickle_suffix, img_suffix, seg_suffix, mean_suffix, std_suffix, neuron_distance_suffix, second_channel, neuron_channel, psf, spacing, vessel_segment_limit, max_pixel_value, n_iter_deconv, grid_size_psf_deconv, sampling, n_cores, filter_radii = False, butter_N = None, butter_Wn = None, butter_btype = None, butter_fs = None, skip_existing=False):
        self.files = files
        self.in_directory = in_directory
        self.img_directory = img_directory
        self.mean_directory = mean_directory
        self.std_directory = std_directory
        self.out_directory = out_directory
        self.pickle_file_suffix = pickle_file_suffix
        self.out_pickle_suffix = out_pickle_suffix
        self.img_suffix = img_suffix
        self.seg_suffix = seg_suffix
        self.mean_suffix = mean_suffix
        self.std_suffix = std_suffix
        self.neuron_distance_suffix = neuron_distance_suffix
        self.second_channel = second_channel
        self.neuron_channel = neuron_channel
        self.psf = psf
        self.spacing = spacing
        self.vessel_segment_limit = vessel_segment_limit
        self.max_pixel_value = max_pixel_value
        self.n_iter_deconv = n_iter_deconv
        self.grid_size_psf_deconv = grid_size_psf_deconv
        self.sampling = sampling
        self.n_cores = n_cores
        self.filter = filter_radii
        self.butter_N = butter_N
        self.butter_Wn = butter_Wn
        self.butter_btype = butter_btype
        self.butter_fs = butter_fs
        self.skip_existing = skip_existing

    def calc_radii_vessels(self, file):

        butter_b, butter_a = butter(self.butter_N, self.butter_Wn, btype=self.butter_btype, analog=False, fs=self.butter_fs)

        if self.skip_existing==False or not exists(sub(self.in_directory, self.out_directory, sub(self.pickle_file_suffix, self.out_pickle_suffix, file))):
            graph = read_gpickle(file)
            if len(graph.edges) < self.vessel_segment_limit:
                img_file = sub(self.in_directory, self.img_directory, sub(self.pickle_file_suffix, self.img_suffix, file))
                seg_file = sub(self.pickle_file_suffix, self.seg_suffix, file)
                mean_file = sub(self.in_directory, self.mean_directory, sub(self.pickle_file_suffix, self.mean_suffix, file))
                std_file = sub(self.in_directory, self.std_directory, sub(self.pickle_file_suffix, self.std_suffix, file))
                img = imread(img_file)
                if self.second_channel:
                    img_ch2 = zoom(swapaxes(img[:, 1, :, :], 0, 2), self.spacing)
                img = zoom(swapaxes(img[:, 0, :, :], 0, 2), self.spacing)
                seg = load(seg_file)
                #mean_img = load(mean_file)
                #std_img = load(std_file)
                seg_dst = distance_transform_edt(seg)
                if self.neuron_channel:
                    nrn_dst = load(sub(self.pickle_file_suffix, self.neuron_distance_suffix, file))
                a, b, c = mgrid[-1 * self.grid_size_psf_deconv // 2:(self.grid_size_psf_deconv // 2 + 1):1,
                                -1 * self.grid_size_psf_deconv // 2:(self.grid_size_psf_deconv // 2 + 1):1,
                                -1 * self.grid_size_psf_deconv // 2:(self.grid_size_psf_deconv // 2 + 1):1]
                abc = dstack([a.flat, b.flat, c.flat])
                mu = array([0, 0, 0])
                sigma = self.psf
                covariance = diag(sigma ** 2)
                d = multivariate_normal.pdf(abc, mean=mu, cov=covariance)
                d = d.reshape((len(a), len(b), len(c)))
                deconv_img = copy(img)
                deconv_img = self.max_pixel_value * richardson_lucy(img / self.max_pixel_value, d,
                                                                    num_iter=self.n_iter_deconv)
                if self.second_channel:
                    deconv_img -= self.max_pixel_value * richardson_lucy(img_ch2 / self.max_pixel_value, d,
                                                                         num_iter=self.n_iter_deconv)
                deconv_img = float32(deconv_img)

                for i in tqdm(range(len(graph.edges))):
                    path = graph[list(graph.edges)[i][0]][list(graph.edges)[i][1]]['pts']
                    _pred_radii = mean(seg_dst[path[::-1, 0], path[::-1, 1], path[::-1, 2]])
                    _pred_radii_max = max(seg_dst[path[::-1, 0], path[::-1, 1], path[::-1, 2]])
                    if _pred_radii == 0:
                        _pred_radii = 1

                    _box_fit = max([int16(_pred_radii_max) + 10, 15])
                    path_smooth = float32(copy(path))
                    for k in range(len(path[0])):
                        path_smooth[:, k] = gaussian_filter1d(float64(path[:, k]), 3, mode='nearest')

                    path_grad = gradient(path_smooth, edge_order=2)[0]
                    res_fwhm = []
                    X = arange(-1 * _box_fit, _box_fit + 1, 1)
                    Y = arange(-1 * _box_fit, _box_fit + 1, 1)
                    x, y = meshgrid(X, Y)
                    x = x.flatten()
                    y = y.flatten()
                    z = zeros(len(x))
                    xy = vstack([x, y, z])

                    def calc_fwhm_path(I):
                        point_grad = path_grad[I]
                        point = path[I]
                        if all(point_grad[0:2] == [0, 0]) and abs(point_grad[2] / point_grad[2]) == 1:
                            rotated = xy.T + point
                        else:
                            rotated = _rotmat(point_grad, xy.T) + point
                        points_img = map_coordinates(deconv_img,
                                                        rotated.T,
                                                        order=3,
                                                        mode='constant')

                        points_img = reshape(points_img, (len(X), len(Y)))
                        points_img = gaussian_filter(points_img, sigma=_pred_radii * .4)

                        _point = array(arange(0, _pred_radii + 20, self.sampling))
                        _zeros = zeros(len(_point))
                        _point = array([_point, _zeros])
                        _centre = closest_node([len(X) // 2 + 1, len(Y) // 2 + 1], peak_local_max(points_img.T))

                        _res = []

                        for deg in arange(0, 360, 10):
                            rot_point = dot(array([[cos(deg2rad(deg)), -1 * sin(deg2rad(deg))],
                                                    [sin(deg2rad(deg)), cos(deg2rad(deg))]]), _point)
                            rot_point[0] = rot_point[0] + _centre[0]
                            rot_point[1] = rot_point[1] + _centre[1]
                            points_vals = map_coordinates(points_img.T,
                                                            rot_point,
                                                            order=3,
                                                            cval=0)
                            points_vals = gaussian_filter1d(points_vals, sigma=_pred_radii * .4 / self.sampling)

                            _ = array(argrelextrema(points_vals , less, order=10))
                            #if _[0,0] == 0:
                            #    _ = array(argrelextrema(points_vals[1:]), less, order=3) + 1

                            if _.shape[1] != 0:
                                points_vals_grad = gradient(points_vals[:max((_[0, 0],int(3.5/self.sampling))) + 3])
                                _ = nanargmin(points_vals_grad)
                                _res.append(_ * self.sampling)

                            else:
                                points_vals_grad = gradient(points_vals)
                                _ = nanargmin(points_vals_grad)
                                _res.append(_ * self.sampling)

                        _res = array(_res)
                        _res = _res[~isnan(_res)]
                        _res = _res[where(_res != 0)]
                        _mean = mean(_res)
                        _std = std(_res)
                        _mask = where(logical_and(_res >= _mean - 2 * _std, _res <= _mean + 2 * _std))
                        _res = _res[_mask]
                        radii = mean(_res)
                        radii_std = std(_res)

                        return radii, radii_std

                    pool = ThreadPool(self.n_cores)
                    _vals, _vals_sigma = zip(*pool.map(calc_fwhm_path, range(len(path))))
                    if self.filter:
                        _vals_unfiltered = copy(_vals)
                        _vals = filtfilt(butter_b, butter_a, _vals, padlen=0)
                    graph[list(graph.edges)[i][0]][list(graph.edges)[i][1]]['radii'] = mean(_vals)
                    graph[list(graph.edges)[i][0]][list(graph.edges)[i][1]]['radii_std'] = std(_vals)
                    graph[list(graph.edges)[i][0]][list(graph.edges)[i][1]]['path_weights'] = _vals
                    if self.filter:
                        graph[list(graph.edges)[i][0]][list(graph.edges)[i][1]]['path_weights_unfiltered'] = _vals_unfiltered
                    graph[list(graph.edges)[i][0]][list(graph.edges)[i][1]]['path_weights_uncertanty'] = _vals_sigma
                    graph[list(graph.edges)[i][0]][list(graph.edges)[i][1]]['end-0z'] = path[0][0]
                    graph[list(graph.edges)[i][0]][list(graph.edges)[i][1]]['end-0y'] = path[0][1]
                    graph[list(graph.edges)[i][0]][list(graph.edges)[i][1]]['end-0x'] = path[0][2]
                    graph[list(graph.edges)[i][0]][list(graph.edges)[i][1]]['end-1z'] = path[-1][0]
                    graph[list(graph.edges)[i][0]][list(graph.edges)[i][1]]['end-1y'] = path[-1][1]
                    graph[list(graph.edges)[i][0]][list(graph.edges)[i][1]]['end-1x'] = path[-1][2]
                    graph[list(graph.edges)[i][0]][list(graph.edges)[i][1]]['mean_depth'] = mean(path[:, 0])
                    graph[list(graph.edges)[i][0]][list(graph.edges)[i][1]]['euclidean-dst'] = sqrt(sum(square(path[-1] - path[0])))
                    if self.neuron_channel:
                        _nrn_dst_vals = nrn_dst[path[::-1, 0], path[::-1, 1], path[::-1, 2]]
                        graph[list(graph.edges)[i][0]][list(graph.edges)[i][1]]['mean_neuron_distance'] = mean(_nrn_dst_vals)
                        graph[list(graph.edges)[i][0]][list(graph.edges)[i][1]]['neuron_distance_std'] = std(_nrn_dst_vals)
                        graph[list(graph.edges)[i][0]][list(graph.edges)[i][1]]['neuron_distance_min'] = min(_nrn_dst_vals)
                write_gpickle(graph, sub(self.in_directory, self.out_directory, sub(self.pickle_file_suffix, self.out_pickle_suffix, file)))
        if 'pool' not in globals():
            pool=5
        del pool
    
    def process_all_files(self):
        for file in tqdm(self.files):
            print(file)
            self.calc_radii_vessels(file)

        




from monai.utils import (
    ensure_tuple
)
from monai.transforms import (
    AsDiscrete,
    EnsureChannelFirstd,
    Compose,
    LoadImaged,
    ScaleIntensityRanged,
    Spacingd,
    EnsureTyped,
    EnsureType,
)
from monai.handlers.utils import from_engine
from monai.networks.nets import UNETR
from monai.networks.layers import Norm
from monai.inferers import sliding_window_inference
from monai.data import ( 
    DataLoader, 
    Dataset, 
    ImageReader
)
from monai.config import (
    KeysCollection, 
    PathLike
)
import torch
import matplotlib.pyplot as plt
from os.path import exists
import os
from pathlib import Path
from re import sub
from skimage.io import imread
from typing import (
    Optional, 
    Union, 
    Sequence,  
    Dict, 
    List
)
from monai.data.utils import is_supported_format
from monai. data.image_reader import (
    _copy_compatible_dict, 
    _stack_images
)
from nibabel.nifti1 import Nifti1Image
from numpy import swapaxes, ndarray, eye, asarray, delete, empty, float16, save
from tqdm import tqdm
from monai.data import DataLoader, Dataset
from torch import no_grad
import pickle
import warnings
warnings.filterwarnings('ignore')



class TIFFReader(ImageReader):
    
    def __init__(self, npz_keys: Optional[KeysCollection] = None, channel_dim: Optional[int] = None, **kwargs):
        super().__init__()
        if npz_keys is not None:
            npz_keys = ensure_tuple(npz_keys)
        self.npz_keys = npz_keys
        self.channel_dim = channel_dim
        self.kwargs = kwargs
    
    def verify_suffix(self, filename: Union[Sequence[PathLike], PathLike]) -> bool:
            """
            Verify whether the specified file or files format is supported by Numpy reader.

            Args:
                filename: file name or a list of file names to read.
                    if a list of files, verify all the suffixes.

            Returns:
                bool: True if the file format is supported, False otherwise.
            """
            suffixes: Sequence[str] = ["tif", "tiff"]
            return is_supported_format(filename, suffixes)

    def read(self, data: Union[Sequence[PathLike], PathLike], **kwargs):
            """
            Read image data from specified file or files, it can read a list of `no-channel` data files
            and stack them together as multi-channels data in `get_data()`.
            Note that the returned object is Numpy array or list of Numpy arrays.

            Args:
                data (Union[Sequence[PathLike], PathLike]): The file name or a list of file names to read.
                **kwargs: Additional arguments for the `numpy.load` API except `allow_pickle`, which will override `self.kwargs` for existing keys.
                    More details about available arguments can be found here: https://numpy.org/doc/stable/reference/generated/numpy.load.html

            Returns:
                Union[numpy.ndarray, List[numpy.ndarray]]: The loaded image data as a Numpy array or a list of Numpy arrays.

            """
            img_: List[Nifti1Image] = []

            filenames: Sequence[PathLike] = ensure_tuple(data)
            kwargs_ = self.kwargs.copy()
            kwargs_.update(kwargs)
            for name in filenames:
                img = imread(name, **kwargs_)
                img = img.astype('float32')
                if len(img.shape)==4:
                    img = swapaxes(img,0,1)
                    img = swapaxes(img,1,3)
                img_.append(img)
            return img_ if len(img_) > 1 else img_[0]
    
    def get_data(self, img):
            """
            Extract data array and meta data from loaded image and return them.
            This function returns two objects, first is numpy array of image data, second is dict of meta data.
            It constructs `affine`, `original_affine`, and `spatial_shape` and stores them in meta dict.
            When loading a list of files, they are stacked together at a new dimension as the first dimension,
            and the meta data of the first image is used to represent the output meta data.

            Args:
                img: a Numpy array loaded from a file or a list of Numpy arrays.

            Returns:
                Tuple: A tuple containing the following two objects:
                    - img_array (numpy.ndarray): Numpy array of image data.
                    - compatible_meta (dict): Dictionary of meta data.

            """
            img_array: List[ndarray] = []
            compatible_meta: Dict = {}
            if isinstance(img, ndarray):
                img = (img,)

            for i in ensure_tuple(img):
                header = {"affine":eye(5),
                         "labels": {"0": "background",
                                    "1": "vessels",
                                    "2": "neurons",}
                         }
                if isinstance(i, ndarray):
                    # if `channel_dim` is None, can not detect the channel dim, use all the dims as spatial_shape
                    spatial_shape = asarray(i.shape)
                    if isinstance(self.channel_dim, int):
                        spatial_shape = delete(spatial_shape, self.channel_dim)
                    header["spatial_shape"] = spatial_shape
                img_array.append(i)
                header["original_channel_dim"] = self.channel_dim if isinstance(self.channel_dim, int) else "no_channel"
                _copy_compatible_dict(header, compatible_meta)

            return _stack_images(img_array, compatible_meta), compatible_meta


def get_model(parameter_file='NOVAS3D_Vessel_and_Neuron_Segmentation/parameters.pickle',
              spatial_dims=3,
              in_channels=2,
              out_channels=3,
              img_size=(128, 128, 128),
              feature_size=16,
              hidden_size=768,
              mlp_dim=3072,
              pos_embed="perceptron",
              res_block=True,
              norm_name="instance",
              gpu = False):
    """
    Get the UNETR model for prediction.

    Args:
        parameter_file (str): Path to the parameter pickle file.
        spatial_dims (int): Number of spatial dimensions.
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        img_size (tuple): Size of the input image.
        feature_size (int): Size of the feature maps.
        hidden_size (int): Size of the hidden layers.
        mlp_dim (int): Dimension of the MLP layers.
        pos_embed (str): Type of positional embedding.
        res_block (bool): Whether to use residual blocks.
        norm_name (str): Name of the normalization layer.

    Returns:
        torch.nn.Module: The UNETR model for prediction.
    """

    # Load parameters from pickle file
    with open(parameter_file, 'rb') as handle:
        params = pickle.load(handle)

    # Set device for model prediction
    device = torch.device("cuda" if (torch.cuda.is_available() and gpu == True) else "cpu")

    # Create UNETR model
    model = UNETR(
        spatial_dims=spatial_dims,
        in_channels=in_channels,
        out_channels=out_channels,
        img_size=img_size,
        feature_size=feature_size,
        hidden_size=hidden_size,
        mlp_dim=mlp_dim,
        pos_embed=pos_embed,
        res_block=res_block,
        norm_name=norm_name,
        dropout_rate=params["dropout"]
    )

    # Move model to device
    model.to(device)

    if (torch.cuda.is_available() and gpu == True):
        # Use DataParallel for multi-GPU training
        model = torch.nn.DataParallel(model)
        # Load pre-trained model weights
        model.load_state_dict(torch.load(
            sub('parameters.pickle', "model.pth",parameter_file)))

    else: 
        from collections import OrderedDict
        state_dict = torch.load(sub('parameters.pickle', "model.pth",parameter_file), map_location='cpu')
        new_state_dict = OrderedDict()
        for k, v in state_dict.items():
            new_state_dict[k.replace("module.", "")] = v
        model.load_state_dict(new_state_dict)
        # Load pre-trained model weights
        #model.load_state_dict(torch.load(
        #    sub('parameters.pickle', "model.pth",parameter_file),
        #    map_location=torch.device('cpu')))
    model.eval()

    for m in model.modules():
        if m.__class__.__name__.startswith('Dropout'):
            m.train()
    return model

def predict(pred_data, 
            num_evals=20,
            num_channels_out=3, 
            model = get_model(), 
            device='cpu',
            roi_size = (128, 128, 128),
            sw_batch_size = 64):
    """
    Perform prediction using a sliding window inference approach.

    Args:
        pred_data (dict): Dictionary containing the input data for prediction.
        num_evals (int, optional): Number of evaluations to perform. Defaults to 20.
        num_channels_out (int, optional): Number of output channels. Defaults to 3.
        model (nn.Module, optional): Model to use for prediction. Defaults to get_model().
        device (str, optional): Device to use for computation. Defaults to 'cpu'.
        roi_size (tuple, optional): Size of the region of interest. Defaults to (128, 128, 128).
        sw_batch_size (int, optional): Batch size for sliding window inference. Defaults to 64.

    Returns:
        numpy.ndarray: Array containing the predicted outputs.
    """
    # Get the shape of the input image
    _shape = pred_data["image"].shape

    # Create an empty array to store the predicted outputs
    pred_array = empty((num_evals,num_channels_out,_shape[2],_shape[3],_shape[4]),dtype=float16)

    # Perform sliding window inference for the specified number of evaluations
    for j in range(num_evals):
        # define softmax for the predicted outputs
        softmax = torch.nn.Softmax(dim=1)
        # Perform sliding window inference
        pred_outputs = sliding_window_inference(
            pred_data["image"],
            roi_size, 
            sw_batch_size, 
            model,
            sw_device=device,
            device=device
        )
        
        # Apply softmax to the predicted outputs
        pred_outputs = softmax(pred_outputs)
        
        # Convert the predicted outputs to numpy array of float16 data type
        pred_outputs = float16(pred_outputs.cpu().detach().numpy())
        
        # Store the predicted outputs in the pred_array
        pred_array[j] = pred_outputs[:]
        del pred_outputs

    return pred_array


def get_pred_transforms(spacing=(1, 1, 0.375), i_min=0, i_max=1024, b_min=0.0, b_max=1.0, clip=True, channel_dim=0):
    """
    Returns a composition of transforms for preprocessing the input image for prediction.

    Args:
        spacing (tuple): Voxel spacing in the input image. Default is (1, 1, 0.375).
        i_min (int): Minimum intensity value for rescaling. Default is 0.
        imax (int): Maximum intensity value for rescaling. Default is 1024.
        b_min (float): Minimum intensity value for scaling. Default is 0.0.
        b_max (float): Maximum intensity value for scaling. Default is 1.0.
        clip (bool): Whether to clip the intensity values. Default is True.
        channel_dim (int): Dimension index for channel. Default is 0.

    Returns:
        pred_transforms (Compose): Composition of transforms for preprocessing the input image.
    """
    # Define a composition of transforms
    pred_transforms = Compose(
        [
            LoadImaged(keys=["image"], reader=TIFFReader(channel_dim=channel_dim)),  # Load the image using TIFFReader
            EnsureChannelFirstd(keys=["image"]),  # Ensure the channel dimension is first
            Spacingd(keys=["image"], pixdim=spacing, mode=("bilinear")),  # Resample the image with the given spacing
            ScaleIntensityRanged(
                keys=["image"], a_min=i_min, a_max=i_max, b_min=b_min, b_max=b_max, clip=clip
            ),  # Rescale the intensity values of the image
            EnsureTyped(keys=["image"]),  # Ensure the image is of the correct data type
        ]
    )
    return pred_transforms

def get_post_transforms():
    """
    Returns the post-processing transforms for the prediction.

    Returns:
        Compose: A composition of post-processing transforms.
    """
    # Define post-processing transforms for prediction
    post_pred = Compose(
        [
            EnsureType(), 
            AsDiscrete(argmax=True, to_onehot=3)
        ]
    )
    return post_pred

class PredictWarped:
    """
    Class for performing prediction using the registered images. Saves the predicted images.

    Args:
        data_dict (dict): A dictionary containing the data for prediction.
        config (dict): A dictionary containing the configuration parameters.
            The configuration parameters include:
                - in_dir (str): The input directory.
                - out_dir (str): The output directory.
                - num_evals (int): The number of evaluations.
                - base_file_extension (str): The base file extension.
                - pred_file_extension (str): The prediction file extension.
                - mean_file_extension (str): The mean file extension.
        parameter_file (str): The file path to the parameter file.
        spatial_dims (int): The number of spatial dimensions.
        in_channels (int): The number of input channels.
        out_channels (int): The number of output channels.
        img_size (tuple): The size of the input image.
        feature_size (int): The size of the features.
        hidden_size (int): The size of the hidden layers.
        mlp_dim (int): The dimension of the MLP layers.
        pos_embed (bool): Whether to use positional embedding.
        res_block (bool): Whether to use residual blocks.
        norm_name (str): The name of the normalization method.
        spacing (tuple): The spacing between voxels. Reciprocal of the voxel size in um to resample to 1 um.
        i_min (float): The minimum intensity value of raw images.
        i_max (float): The maximum intensity value of raw images.
        b_min (float): The minimum background value after normalization.
        b_max (float): The maximum background value after noramlization.
        clip (bool): Whether to clip the values.
        channel_dim (int): The dimension index of the colour channel.
        gpu (bool): Whether to use the GPU.
        skip_existing (bool): Whether to skip images that have already been predicted.
    
    Methods:
        get_model(): Retrieves the model for prediction.
        get_pred_transforms(): Retrieves the prediction transforms.
        dataset_creation(): Creates the dataset for prediction.
        prediction(): Performs the prediction.

    """

    def __init__(self, data_dict, config, parameter_file, spatial_dims, in_channels, out_channels, img_size, feature_size, hidden_size, mlp_dim, pos_embed, res_block, norm_name,spacing, i_min, i_max, b_min, b_max, clip, channel_dim, gpu=False, skip_existing=False):
        self.data_dict = data_dict
        self.config = config
        self.parameter_file = parameter_file
        self.spatial_dims = spatial_dims
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.img_size = img_size
        self.feature_size = feature_size
        self.hidden_size = hidden_size
        self.mlp_dim = mlp_dim
        self.pos_embed = pos_embed
        self.res_block = res_block
        self.norm_name = norm_name
        self.spacing = spacing
        self.i_min = i_min
        self.i_max = i_max
        self.b_min = b_min
        self.b_max = b_max
        self.clip = clip
        self.channel_dim = channel_dim
        self.gpu = gpu
        self.skip_existing = skip_existing

    def get_model(self):
        """
        Get the model for prediction.

        Returns:
            model: The model for prediction.
        """
        # Get the model for prediction
        model = get_model(self.parameter_file, 
                          self.spatial_dims,
                          self.in_channels, 
                          self.out_channels, 
                          self.img_size, 
                          self.feature_size, 
                          self.hidden_size, 
                          self.mlp_dim, 
                          self.pos_embed, 
                          self.res_block, 
                          self.norm_name,
                          self.gpu)
        return model

    def get_pred_transforms(self):
        """
        Get the prediction transforms.

        Returns:
            pred_transforms (list): List of prediction transforms.
        """
        # Get the prediction transforms
        pred_transforms = get_pred_transforms(self.spacing, 
                                              self.i_min, 
                                              self.i_max, 
                                              self.b_min, 
                                              self.b_max, 
                                              self.clip, 
                                              self.channel_dim)
        return pred_transforms

    def dataset_creation(self):
        """
        Create the dataset for prediction.

        Returns:
            pred_ds (Dataset): The dataset for prediction.
        """
        # Create the dataset for prediction
        data_dicts = self.data_dict
        pred_transforms = get_pred_transforms(self.spacing, self.i_min, self.i_max, self.b_min, self.b_max, self.clip, self.channel_dim)
        pred_ds = Dataset(data=data_dicts, transform=pred_transforms)
        self.assertEqual(len(pred_ds), 3)
        return pred_ds

    def prediction(self):
        """
        Perform prediction using the provided model and saves the data.

        Returns:
            None
        """
        # Perform the prediction
        model = get_model(self.parameter_file, self.spatial_dims, self.in_channels, self.out_channels, self.img_size, self.feature_size, self.hidden_size, self.mlp_dim, self.pos_embed, self.res_block, self.norm_name)
        pred_transforms = get_pred_transforms(self.spacing, self.i_min, self.i_max, self.b_min, self.b_max, self.clip, self.channel_dim)
        pred_ds = Dataset(data=self.data_dict, transform=pred_transforms)
        pred_loader = DataLoader(pred_ds, batch_size=1, shuffle=False)
        config = self.config
        with no_grad():
            for i, pred_data in tqdm(enumerate(pred_loader)):
                if self.skip_existing==False:
                    new_file_name = sub(config["in_dir"], config["out_dir"], self.data_dict[i]["image"])
                    pred_array = predict(pred_data, num_evals=config["num_evals"], model=model)
                    #self.assertIsNotNone(pred_array)
                    mean = float16(pred_array.mean(axis=0))
                    save(sub(config["base_file_extension"], config['mean_file_extension'], new_file_name), mean)
                    save(sub(config["base_file_extension"], config["pred_file_extension"], new_file_name), pred_array)
                    
                elif not exists(sub(config["base_file_extension"], config["pred_file_extension"], sub(config["in_dir"], config["out_dir"], self.data_dict[i]["image"]))):
                    new_file_name = sub(config["in_dir"], config["out_dir"], self.data_dict[i]["image"])
                    pred_array = predict(pred_data, num_evals=config["num_evals"], model=model)
                    #self.assertIsNotNone(pred_array)
                    mean = float16(pred_array.mean(axis=0))
                    save(sub(config["base_file_extension"], config['mean_file_extension'], new_file_name), mean)
                    save(sub(config["base_file_extension"], config["pred_file_extension"], new_file_name), pred_array)